/*
 * Copyright (c) 2016-2019, NVIDIA CORPORATION. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of NVIDIA CORPORATION nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <poll.h>

#include "NvJpegEncoder.h"
#include "NvVideoConverter.h"
#include "NvEglRenderer.h"
#include "NvUtils.h"
#include "NvCudaProc.h"
#include "nvbuf_utils.h"

#include "camera_v4l2_cuda.h"

#define MJPEG_EOS_SEARCH_SIZE 4096

static bool quit = false;

using namespace std;

/*
 * 打印用法信息
 */
static void
print_usage(void)
{
    printf("\n\t用法: camera_v4l2_cuda [OPTIONS]\n\n"
           "\t示例: \n"
           "\t./camera_v4l2_cuda -d /dev/video0 -s 640x480 -f YUYV -n 30 -c\n\n"
           "\t支持参数:\n"
           "\t-d\t\t设置V4l2视频设备节点\n"
           "\t-s\t\t设置视频设备的输出分辨率\n"
           "\t-f\t\t视频设备的输出像素格式 (只支持 YUYV/YVYU/UYVY/VYUY/GREY/MJPEG)\n"
           "\t-r\t\t设置渲染器帧速率 (默认30帧/秒)\n"
           "\t-n\t\t保存VIC处理前的第n帧\n"
           "\t-c\t\t支持CUDA aglorithm (draw a black box in the upper left corner)\n"
           "\t-v\t\t启用详细信息\n"
           "\t-h\t\t打印这个用法\n\n"
           "\tNOTE: 它无限运行，直到你终止它，快捷键 <ctrl+c>\n");
}

/*
 * 解析并更新参数
 */
static bool
parse_cmdline(context_t * ctx, int argc, char **argv)
{
    int c;

    if (argc < 2)
    {
        print_usage();
        exit(EXIT_SUCCESS);
    }

    while ((c = getopt(argc, argv, "d:s:f:r:n:cvh")) != -1)
    {
        switch (c)
        {
            case 'd':   // 设置摄像头设备
                ctx->cam_devname = optarg;
                break;

            case 's':   // 设置分辨率
                if (sscanf(optarg, "%dx%d",
                            &ctx->cam_w, &ctx->cam_h) != 2)
                {
                    print_usage();
                    return false;
                }
                break;

            case 'f':   // 设置图像采集格式
                if (strcmp(optarg, "YUYV") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_YUYV;
                else if (strcmp(optarg, "YVYU") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_YVYU;
                else if (strcmp(optarg, "VYUY") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_VYUY;
                else if (strcmp(optarg, "UYVY") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_UYVY;
                else if (strcmp(optarg, "GREY") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_GREY;
                else if (strcmp(optarg, "MJPEG") == 0)
                    ctx->cam_pixfmt = V4L2_PIX_FMT_MJPEG;
                else
                {
                    print_usage();
                    return false;
                }
                sprintf(ctx->cam_file, "camera.%s", optarg);
                break;

            case 'r':   // 设置渲染帧速率
                ctx->fps = strtol(optarg, NULL, 10);
                break;

            case 'n':   // 保存VIC处理前的第n帧
                ctx->save_n_frame = strtol(optarg, NULL, 10);
                break;

            case 'c':   // 启用cuda
                ctx->enable_cuda = true;
                break;

            case 'v':   // 启用log
                ctx->enable_verbose = true;
                break;

            case 'h':   // 打印用法
                print_usage();
                exit(EXIT_SUCCESS);
                break;

            default:
                print_usage();
                return false;
        }
    }

    return true;
}

static void
set_defaults(context_t * ctx)
{
    memset(ctx, 0, sizeof(context_t));

    ctx->cam_devname = "/dev/video0";       // video设备
    ctx->cam_fd = -1;                       // 摄像头描述符
    ctx->cam_pixfmt = V4L2_PIX_FMT_YUYV;    // 像素格式
    ctx->cam_w = 640;
    ctx->cam_h = 480;
    ctx->frame = 0;
    ctx->save_n_frame = 0;

    ctx->g_buff = NULL;                     // 缓冲区
    ctx->capture_dmabuf = true;             // 捕获dma缓冲区
    ctx->renderer = NULL;                   // EGL渲染器
    ctx->fps = 30;

    ctx->enable_cuda = false;
    ctx->egl_image = NULL;
    ctx->egl_display = EGL_NO_DISPLAY;

    ctx->enable_verbose = false;
}

static nv_color_fmt nvcolor_fmt[] =
{
    /* TODO: add more pixel format mapping */
    {V4L2_PIX_FMT_UYVY, NvBufferColorFormat_UYVY},
    {V4L2_PIX_FMT_VYUY, NvBufferColorFormat_VYUY},
    {V4L2_PIX_FMT_YUYV, NvBufferColorFormat_YUYV},
    {V4L2_PIX_FMT_YVYU, NvBufferColorFormat_YVYU},
    {V4L2_PIX_FMT_GREY, NvBufferColorFormat_GRAY8},
    {V4L2_PIX_FMT_YUV420M, NvBufferColorFormat_YUV420},
};

// 取得缓冲区对于v4l2的图像格式
static NvBufferColorFormat
get_nvbuff_color_fmt(unsigned int v4l2_pixfmt)
{
    unsigned i;

    for (i = 0; i < sizeof(nvcolor_fmt) / sizeof(nvcolor_fmt[0]); i++)
    {
        if (v4l2_pixfmt == nvcolor_fmt[i].v4l2_pixfmt)
            return nvcolor_fmt[i].nvbuff_color;
    }

    return NvBufferColorFormat_Invalid;
}

static bool
save_frame_to_file(context_t * ctx, struct v4l2_buffer * buf)
{
    int file;

    file = open(ctx->cam_file, O_CREAT | O_WRONLY | O_APPEND | O_TRUNC,
            S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

    if (-1 == file)
        ERROR_RETURN("无法打开文件进行框架保存");

    if (-1 == write(file, ctx->g_buff[buf->index].start,
                ctx->g_buff[buf->index].size))
    {
        close(file);
        ERROR_RETURN("Failed to write frame into file");
    }

    close(file);

    return true;
}

static bool
nvbuff_do_clearchroma (int dmabuf_fd)
{
    NvBufferParams params = {0};
    void *sBaseAddr[3] = {NULL};
    int ret = 0;
    int size;
    unsigned i;

    ret = NvBufferGetParams (dmabuf_fd, &params);// 获取缓冲区参数
    if (ret != 0)
        ERROR_RETURN("%s: NvBufferGetParams 失败 \n", __func__);

    for (i = 1; i < params.num_planes; i++)
    {
        /* 获取平面的内存映射虚拟地址 */
        ret = NvBufferMemMap (dmabuf_fd, i, NvBufferMem_Read_Write, &sBaseAddr[i]);
        if (ret != 0)
            ERROR_RETURN("%s: NvBufferMemMap 失败 \n", __func__);

        /* 由于数据来自VIC，因此同步设备缓存以进行CPU访问 */
        ret = NvBufferMemSyncForCpu (dmabuf_fd, i, &sBaseAddr[i]);// 同步CPU的硬件存储器高速缓存
        if (ret != 0)
            ERROR_RETURN("%s: NvBufferMemSyncForCpu 失败 \n", __func__);

        size = params.height[i] * params.pitch[i];// 高度 * 间距
        memset (sBaseAddr[i], 0x80, size);

        /* 同步CPU缓存，以便VIC访问，因为数据来自CPU */
        ret = NvBufferMemSyncForDevice (dmabuf_fd, i, &sBaseAddr[i]);// 同步设备的硬件内存缓存
        if (ret != 0)
            ERROR_RETURN("%s: NvBufferMemSyncForDevice 失败 \n", __func__);

        /* 取消映射平面的映射虚拟地址 */
        ret = NvBufferMemUnMap (dmabuf_fd, i, &sBaseAddr[i]);
        if (ret != 0)
            ERROR_RETURN("%s: NvBufferMemUnMap 失败 \n", __func__);
    }

    return true;
}

// 初始化摄像机装置
static bool
camera_initialize(context_t * ctx)
{
    struct v4l2_format fmt;

    /* 开式摄影机 */
    ctx->cam_fd = open(ctx->cam_devname, O_RDWR);
    if (ctx->cam_fd == -1)
    {
        ERROR_RETURN("未能打开摄像机装置 %s: %s (%d)", ctx->cam_devname, strerror(errno), errno);
    }
    else
    {
        INFO("摄像机装置打开成功。");
    }

    /* 设置摄像机输出格式 */
    memset(&fmt, 0, sizeof(fmt));
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;// 视频捕捉模式
    fmt.fmt.pix.width = ctx->cam_w;
    fmt.fmt.pix.height = ctx->cam_h;
    INFO("采集图像分辨率：%dx%d", fmt.fmt.pix.width, fmt.fmt.pix.height);
    switch (fmt.fmt.pix.pixelformat = ctx->cam_pixfmt)
    {
        case V4L2_PIX_FMT_YUYV:
            INFO("采集像素格式：%s", "YUYV");
            break;

        case V4L2_PIX_FMT_YVYU:
            INFO("采集像素格式：%s", "YVYU");
            break;

        case V4L2_PIX_FMT_VYUY:
            INFO("采集像素格式：%s", "VYUY");
            break;

        case V4L2_PIX_FMT_UYVY:
            INFO("采集像素格式：%s", "UYVY");
            break;

        case V4L2_PIX_FMT_GREY:
            INFO("采集像素格式：%s", "GREY");
            break;

        case V4L2_PIX_FMT_MJPEG:
            INFO("采集像素格式：%s", "MJPEG");
            break;

        default:
            INFO("不支持的采集像素格式");
            break;
    }

    fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    if (ioctl(ctx->cam_fd, VIDIOC_S_FMT, &fmt) < 0)
    {
        ERROR_RETURN("未能设置相机输出格式: %s (%d)", strerror(errno), errno);
    }

    /* 在不支持的情况下,获取真实支持的格式 */
    memset(&fmt, 0, sizeof fmt);
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(ctx->cam_fd, VIDIOC_G_FMT, &fmt) < 0)
    {
        ERROR_RETURN("未能获得相机输出格式: %s (%d)", strerror(errno), errno);
    }

    if (fmt.fmt.pix.width != ctx->cam_w ||
        fmt.fmt.pix.height != ctx->cam_h ||
        fmt.fmt.pix.pixelformat != ctx->cam_pixfmt)
    {
        WARN("不支持所需的格式");
        ctx->cam_w = fmt.fmt.pix.width;
        ctx->cam_h = fmt.fmt.pix.height;
        ctx->cam_pixfmt =fmt.fmt.pix.pixelformat;
    }

    struct v4l2_streamparm streamparm;
    memset (&streamparm, 0x00, sizeof (struct v4l2_streamparm));
    streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    ioctl (ctx->cam_fd, VIDIOC_G_PARM, &streamparm);

    INFO("照相机格式: (%d x %d)  stride: %d, 图像大小: %d, frate: %u / %u",
        fmt.fmt.pix.width,
        fmt.fmt.pix.height,
        fmt.fmt.pix.bytesperline,
        fmt.fmt.pix.sizeimage,
        streamparm.parm.capture.timeperframe.denominator,
        streamparm.parm.capture.timeperframe.numerator);

    return true;
}

static bool
display_initialize(context_t * ctx)
{
    /* 创建EGL渲染 */
    ctx->renderer = NvEglRenderer::createEglRenderer("renderer0", ctx->cam_w, ctx->cam_h, 0, 0);
    if (!ctx->renderer)
    {
        ERROR_RETURN("没有成功创建EGL渲染");
    }

    ctx->renderer->setFPS(ctx->fps);

    if (ctx->enable_cuda)
    {
        /* 获取默认EGL显示 */
        ctx->egl_display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
        if (ctx->egl_display == EGL_NO_DISPLAY)
            ERROR_RETURN("没有得到EGL显示连接");

        /* 初始化EGL显示连接 */
        if (!eglInitialize(ctx->egl_display, NULL, NULL))
            ERROR_RETURN("没有初始化EGL显示连接");
    }

    return true;
}

// 初始化摄像机装置及显示渲染器
static bool
init_components(context_t * ctx)
{
    if (!camera_initialize(ctx))
        ERROR_RETURN("未初始化摄像机装置");

    if (!display_initialize(ctx))
        ERROR_RETURN("未初始化显示");

    INFO("成功初始化v4l2组件");
    return true;
}

static bool
request_camera_buff(context_t *ctx)
{
    /* 请求摄像机v4l2缓冲区 */
    struct v4l2_requestbuffers rb;
    memset(&rb, 0, sizeof(rb));
    rb.count = V4L2_BUFFERS_NUM;
    rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rb.memory = V4L2_MEMORY_DMABUF;
    if (ioctl(ctx->cam_fd, VIDIOC_REQBUFS, &rb) < 0)
        ERROR_RETURN("请求v4l2缓冲区失败: %s (%d)", strerror(errno), errno);

    if (rb.count != V4L2_BUFFERS_NUM)
        ERROR_RETURN("V4l2缓冲区号不理想");

    for (unsigned int index = 0; index < V4L2_BUFFERS_NUM; index++)
    {
        struct v4l2_buffer buf;

        /* 查询摄像机v4l2 buf长度 */
        memset(&buf, 0, sizeof buf);
        buf.index = index;
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_DMABUF;

        /* 数据缓存转换成物理地址 */
        if (ioctl(ctx->cam_fd, VIDIOC_QUERYBUF, &buf) < 0)
            ERROR_RETURN("无法查询buff: %s (%d)",
                    strerror(errno), errno);

        /* TODO: 添加对多平面机的支持将空的v4l2 buff入摄像机捕获平面 */
        buf.m.fd = (unsigned long)ctx->g_buff[index].dmabuff_fd;
        if (buf.length != ctx->g_buff[index].size)
        {
            WARN("相机v4l2 buf长度不正确");
            ctx->g_buff[index].size = buf.length;
        }

        /* 把数据从缓存中读取出来 */
        if (ioctl(ctx->cam_fd, VIDIOC_QBUF, &buf) < 0)
            ERROR_RETURN("无法读入缓冲区: %s (%d)",
                    strerror(errno), errno);
    }

    return true;
}

static bool
request_camera_buff_mmap(context_t *ctx)
{
    /* 请求相机v4l2缓冲器 */
    struct v4l2_requestbuffers rb;
    memset(&rb, 0, sizeof(rb));
    rb.count = V4L2_BUFFERS_NUM;
    rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rb.memory = V4L2_MEMORY_MMAP;
    if (ioctl(ctx->cam_fd, VIDIOC_REQBUFS, &rb) < 0)
        ERROR_RETURN("未能请求v4l2缓冲区: %s (%d)", strerror(errno), errno);
    if (rb.count != V4L2_BUFFERS_NUM)
        ERROR_RETURN("V4l2缓冲区数量非想要的那样");

    for (unsigned int index = 0; index < V4L2_BUFFERS_NUM; index++)
    {
        struct v4l2_buffer buf;

        /* 查询相机v4l2 buf长度 */
        memset(&buf, 0, sizeof buf);
        buf.index = index;
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        buf.memory = V4L2_MEMORY_MMAP;
        if (ioctl(ctx->cam_fd, VIDIOC_QUERYBUF, &buf) < 0)
            ERROR_RETURN("未能查询大小: %s (%d)", strerror(errno), errno);

        ctx->g_buff[index].size = buf.length;
        ctx->g_buff[index].start = (unsigned char *)
            mmap (NULL /* start anywhere */,
                    buf.length,
                    PROT_READ | PROT_WRITE /* required */,
                    MAP_SHARED /* recommended */,
                    ctx->cam_fd, buf.m.offset);
        if (MAP_FAILED == ctx->g_buff[index].start)
            ERROR_RETURN("无法映射缓冲区");

        if (ioctl(ctx->cam_fd, VIDIOC_QBUF, &buf) < 0)
            ERROR_RETURN("未能锁定缓冲区: %s (%d)", strerror(errno), errno);
    }

    return true;
}

static bool
prepare_buffers_mjpeg(context_t * ctx)
{
    NvBufferCreateParams input_params = {0};

    /* 分配全局缓冲区上下文 */
    ctx->g_buff = (nv_buffer *)malloc(V4L2_BUFFERS_NUM * sizeof(nv_buffer));
    if (ctx->g_buff == NULL)
    {
        ERROR_RETURN("未能分配全局缓冲区上下文");
    }
    memset(ctx->g_buff, 0, V4L2_BUFFERS_NUM * sizeof(nv_buffer));

    input_params.payloadType = NvBufferPayload_SurfArray;
    input_params.width = ctx->cam_w;
    input_params.height = ctx->cam_h;
    input_params.layout = NvBufferLayout_Pitch;
    input_params.colorFormat = get_nvbuff_color_fmt(V4L2_PIX_FMT_YUV420M);
    input_params.nvbuf_tag = NvBufferTag_NONE;

    /* 创建呈现缓冲区 */
    if (-1 == NvBufferCreateEx(&ctx->render_dmabuf_fd, &input_params))
        ERROR_RETURN("未能创建NvBuffer");

    ctx->capture_dmabuf = false;
    if (!request_camera_buff_mmap(ctx))
        ERROR_RETURN("没有设置相机缓冲区");

    INFO("成功编写mjpeg缓冲区");
    return true;
}

static bool
prepare_buffers(context_t * ctx)
{
    // 创建硬件缓冲区的输入参数
    NvBufferCreateParams input_params = {0};

    /* 分配全局缓冲区上下文 */
    ctx->g_buff = (nv_buffer *)malloc(V4L2_BUFFERS_NUM * sizeof(nv_buffer));
    if (ctx->g_buff == NULL)
        ERROR_RETURN("未能分配全局缓冲区上下文");

    // 设置缓冲区的有效负载类型
    input_params.payloadType = NvBufferPayload_SurfArray;// 带硬件存储器句柄的缓冲有效载荷，用于一组平面
                            // NvBufferPayload_MemHandle; // 具有特定内存大小的HW内存句柄的缓冲区有效负载
    input_params.width = ctx->cam_w;
    input_params.height = ctx->cam_h;
    // 缓冲区布局
    input_params.layout = NvBufferLayout_Pitch;// 间距布局
                        // NvBufferLayout_BlockLinear; // 线性块布局

    /* 创建缓冲区并准备摄像头 */
    for (unsigned int index = 0; index < V4L2_BUFFERS_NUM; index++)
    {
        int fd;
        // 硬件缓冲区的参数
        NvBufferParams params = {0};

        // 缓冲区的颜色格式
        input_params.colorFormat = get_nvbuff_color_fmt(ctx->cam_pixfmt);
        // 标签与缓冲区关联
        input_params.nvbuf_tag = NvBufferTag_CAMERA;// 相机标签

        /* 分配硬件缓冲区 */
        if (-1 == NvBufferCreateEx(&fd, &input_params))
            ERROR_RETURN("未能创建缓冲区");

        ctx->g_buff[index].dmabuff_fd = fd;

        /* 获取缓冲区参数 */
        if (-1 == NvBufferGetParams(fd, &params))
            ERROR_RETURN("未能获得缓冲区参数");

        if (ctx->cam_pixfmt == V4L2_PIX_FMT_GREY &&
            params.pitch[0] != params.width[0])// 硬件缓冲区每个平面的间距
                ctx->capture_dmabuf = false;

        /* TODO: 添加多平面支持目前只支持yuvv插入单平面 */
        if (ctx->capture_dmabuf) {
            /* 获取平面的内存映射虚拟地址 */
            if (-1 == NvBufferMemMap(ctx->g_buff[index].dmabuff_fd, 0, NvBufferMem_Read_Write,
                        (void**)&ctx->g_buff[index].start))
                ERROR_RETURN("无法映射缓冲区");
        }
    }

    // 缓冲区的颜色格式
    input_params.colorFormat = get_nvbuff_color_fmt(V4L2_PIX_FMT_YUV420M);
    // 标签与缓冲区关联
    input_params.nvbuf_tag = NvBufferTag_NONE;// 无标记
    /* 分配硬件缓冲区 */
    if (-1 == NvBufferCreateEx(&ctx->render_dmabuf_fd, &input_params))
        ERROR_RETURN("未能创建缓冲区");

    if (ctx->capture_dmabuf) {
        if (!request_camera_buff(ctx))
            ERROR_RETURN("没有设置相机抛光");
    } else {
        if (!request_camera_buff_mmap(ctx))
            ERROR_RETURN("没有设置相机抛光");
    }

    INFO("成功地准备流缓冲区");
    return true;
}

static bool
start_stream(context_t * ctx)
{
    enum v4l2_buf_type type;

    /* 开始v4l2流 */
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    /* 开始视频显示函数 */
    if (ioctl(ctx->cam_fd, VIDIOC_STREAMON, &type) < 0)
        ERROR_RETURN("未能开始流: %s (%d)", strerror(errno), errno);

    usleep(200);

    INFO("摄像机视频流 ...");
    return true;
}

static void
signal_handle(int signum)
{
    printf("由于用户退出命令而退出!\n");
    quit = true;
}

// cuda后续处理
static bool
cuda_postprocess(context_t *ctx, int fd)
{
    if (ctx->enable_cuda)
    {
        /* 从dmabuf fd创建EGLImage实例 */
        ctx->egl_image = NvEGLImageFromFd(ctx->egl_display, fd);
        if (ctx->egl_image == NULL)
            ERROR_RETURN("无法将 dmabuf fd (0x%X) 映射到 EGLImage",
                    ctx->render_dmabuf_fd);

        /* 将此钩在此egl_image上的缓冲区传递给CUDA以进行CUDA处理-在框架上绘制一个矩形 */
        HandleEGLImage(&ctx->egl_image);

        /* 销毁EGLImage */
        NvDestroyEGLImage(ctx->egl_display, ctx->egl_image);
        ctx->egl_image = NULL;
    }

    return true;
}

/**
 * Callback function called after capture plane dqbuffer of NvVideoConverter class.
 * See NvV4l2ElementPlane::dqThread() in sample/common/class/NvV4l2ElementPlane.cpp
 * for details.
 *
 * @param v4l2_buf       : dequeued v4l2 buffer
 * @param buffer         : NvBuffer associated with the dequeued v4l2 buffer
 * @param shared_buffer  : Shared NvBuffer if the queued buffer is shared with
 *                         other elements. Can be NULL.
 * @param arg            : private data set by NvV4l2ElementPlane::startDQThread()
 *
 * @return               : true for success, false for failure (will stop DQThread)
 */
static bool
conv_capture_dqbuf_thread_callback(struct v4l2_buffer *v4l2_buf,
                                   NvBuffer * buffer, NvBuffer * shared_buffer,
                                   void *arg)
{
    context_t *ctx = (context_t *) arg;
    unsigned long out_buf_size = ctx->cam_w * ctx->cam_h * 3 / 2;
    unsigned char *out_buf = new unsigned char[out_buf_size];
    int iterator_num = PERF_LOOP;
    int ret;
    uint64_t i = 0;
    char *out_file_path = NULL;
    sprintf(out_file_path, "%d.jpg", i++);

    INFO("文件名为：%s", out_file_path);

    ctx->out_file = new ofstream(out_file_path);
    if (!ctx->out_file->is_open())
    {
        ERROR_RETURN("无法打开输出文件");
    }

    if (!v4l2_buf)
    {
        delete[] out_buf;
        ERROR_RETURN("无法从转换捕获平面出队缓冲区");
    }

    if (v4l2_buf->m.planes[0].bytesused > 0)
    {
        for (int i = 0; i < iterator_num; ++i)
        {
            /* 从硬件缓冲存储器的文件描述符（FD）编码JPEG图像。*/
            ret = 
                ctx->jpegenc->encodeFromFd (buffer->planes[0].fd, 
                                            JCS_YCbCr, 
                                            &out_buf,
                                            out_buf_size, 
                                            75);// 图像质量
            if (ret < 0)
            {
                ERROR_RETURN("从fd编码时出错");
            }
        }
        if (ret >= 0)
        {
            ctx->out_file->write((char *) out_buf, out_buf_size);
        }
    }

    delete[] out_buf;
    delete ctx->out_file;
    free(out_file_path);

    return false;
}

bool
save_to_jpg(context_t * ctx, struct v4l2_buffer* v4l2_buf)
{
    int ret = 0;
#if 1

    /* 创建一个名为comp_name的新JPEG编码器 */
    ctx->jpegenc = NvJPEGEncoder::createJPEGEncoder("jpenenc");
    if (!ctx->jpegenc)
    {
        ERROR_RETURN("无法创建Jpeg编码器");
    }

    /* 启用元素分析:基准编码器性能 */
    ctx->jpegenc->enableProfiling();

#if 0
    /* 设置JPEG编码器使用的裁剪矩形 */
    //ctx->jpegenc->setCropRect(0, 0, ctx.cam_w, ctx.cam_h);

    if(ctx->scaled_encode)
    {
        /* 设置缩放参数，通过该缩放参数需要缩放和编码图像。 */
        //ctx->jpegenc->setScaledEncodeParams(ctx.cam_w, ctx.cam_h);
    }
#endif

    /* 创建一个V4L2视频转换器对象 */
    ctx->conv = NvVideoConverter::createVideoConverter("conv");
    if (!ctx->conv)
    {
        ERROR_RETURN("无法创建视频转换器");
    }

    /* 设置转换输出平面格式 */
    ret = 
        ctx->conv->setOutputPlaneFormat (V4L2_PIX_FMT_YUV420M,          // 原始V4L2像素格式之一
                                        ctx->cam_w,                     // 输出缓冲区的宽度（以像素为单位）
                                        ctx->cam_h,                     // 输出缓冲区的高度（以像素为单位）
                                        V4L2_NV_BUFFER_LAYOUT_PITCH);   // 平面中的缓冲区布局。间距线性布局
    if (ret < 0)
    {
        ERROR_RETURN("无法为转换设置输出平面格式");
    }

    /* 设置转换捕获平面格式，YUV420或NV12 */
    ret =
        ctx->conv->setCapturePlaneFormat(ctx->cam_pixfmt,
                                        ctx->cam_w, ctx->cam_h,
                                        V4L2_NV_BUFFER_LAYOUT_PITCH);//V4L2_NV_BUFFER_LAYOUT_BLOCKLINEAR);// 块线性布局
    if (ret < 0)
    {
        ERROR_RETURN("无法为转换设置捕获平面格式");
    }

    /* REQBUF，EXPORT和MAP转换输出平面缓冲区 */
    ret = ctx->conv->output_plane.setupPlane(V4L2_MEMORY_MMAP, 1, true, false);
    if (ret < 0)
    {
        ERROR_RETURN("设置转换输出平面时出错");
    }

    ret = ctx->conv->capture_plane.setupPlane(V4L2_MEMORY_MMAP, 1, true, false);
    if (ret < 0)
    {
        ERROR_RETURN("设置转换捕获平面时出错");
    }

    /* 开始转换输出平面流 */
    ret = ctx->conv->output_plane.setStreamStatus(true);
    if (ret < 0)
    {
        ERROR_RETURN("转换输出平面流中的错误");
    }

    /* 开始转换捕获平面流 */
    ret = ctx->conv->capture_plane.setStreamStatus(true);
    if (ret < 0)
    {
        ERROR_RETURN("转换时捕获平面流中的错误");
    }

    /* 在conv捕获平面上注册出队线程的回调，此回调会将YUV420或NV12图像编码为JPEG并写入文件系统。*/
    ctx->conv->capture_plane.setDQThreadCallback(conv_capture_dqbuf_thread_callback);

    // 启动线程以使conv捕获平面上的缓冲区出队
    ctx->conv->capture_plane.startDQThread(&ctx);

    /* 排队所有空的conv捕获平面缓冲区，实际上，在这种情况下，将排队1个缓冲区。*/
    INFO("平面缓冲区数量为：%d", ctx->conv->capture_plane.getNumBuffers());
    for (uint32_t i = 0; i < ctx->conv->capture_plane.getNumBuffers(); i++)
    {
        struct v4l2_buffer capture_v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];

        memset(&capture_v4l2_buf, 0, sizeof(capture_v4l2_buf));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        capture_v4l2_buf.index = i;
        capture_v4l2_buf.m.planes = planes;

        /* 在平面上将缓冲区排队 */
        ret = ctx->conv->capture_plane.qBuffer(capture_v4l2_buf, NULL);
        if (ret < 0)
        {
            ERROR_RETURN("在conv捕获平面上排队缓冲区时出错");
        }
    }
#endif
    /**
     * 将YUV420图像读取到conv输出平面缓冲区并排队，以便conv可以开始处理。
     */
    {
        //struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];
        /* 获取NvBuffer索引为n的对象 */
        //NvBuffer *buffer = ctx->conv->output_plane.getNthBuffer(0);

        memset(v4l2_buf, 0, sizeof(struct v4l2_buffer));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        v4l2_buf->index = 0;
        v4l2_buf->m.planes = planes;

        v4l2_buf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (ctx->capture_dmabuf)
            v4l2_buf->memory = V4L2_MEMORY_DMABUF;
        else
            v4l2_buf->memory = V4L2_MEMORY_MMAP;

        if (ioctl(ctx->cam_fd, VIDIOC_DQBUF, v4l2_buf) < 0)
            ERROR_RETURN("无法从相机读取数据: %s (%d)", strerror(errno), errno);
        /*
        if (read_video_frame(ctx.in_file, buffer) < 0)
        {
            cerr << "无法从文件读取完整的帧" << endl;
            v4l2_buf->m.planes[0].bytesused = 0;
        }
        //*/

#if 1
    INFO("start qbuffer");
        ret = ctx->conv->output_plane.qBuffer(*v4l2_buf, NULL);
        if (ret < 0)
        {
            ERROR_RETURN("在conv输出平面上排队缓冲区时出错");
        }
    }
    INFO("qbuffer OK");

    /* 等待直到conv上的所有捕获平面缓冲区出队 */
    ctx->conv->capture_plane.waitForDQThread(2000);
#endif
}

static bool
start_capture(context_t * ctx)
{
    uint64_t i = 1;
    struct sigaction sig_action;
    struct pollfd fds[1];
    // 保留缓冲区转换函数的参数
    NvBufferTransformParams transParams;

    /* 注册一个shuwdown处理程序,以确保在用户类型和控制下保持干净的关闭 <ctrl+c> */
    sig_action.sa_handler = signal_handle;
    /* 初始化信号集合为空 */
    sigemptyset(&sig_action.sa_mask);
    sig_action.sa_flags = 0;
    /* 安装和登记信号的处理 */
    sigaction(SIGINT, &sig_action, NULL);

    if (ctx->cam_pixfmt == V4L2_PIX_FMT_MJPEG)
        /* 创建一个名为comp_name的新JPEG解码器 */
        ctx->jpegdec = NvJPEGDecoder::createJPEGDecoder("jpegdec");

    /* 初始化缓冲区变换器 */
    memset(&transParams, 0, sizeof(transParams));
    transParams.transform_flag = NVBUFFER_TRANSFORM_FILTER;// 转换标志以设置过滤器类型
    transParams.transform_filter = NvBufferTransform_Filter_Smart;// 智能地转换过滤器

    /* 启用元素分析 */
    ctx->renderer->enableProfiling();

    fds[0].fd = ctx->cam_fd;
    fds[0].events = POLLIN;// 等待的事件
    /* 等待摄像头事件超时 = 1000毫秒 */
    while (poll(fds, 1, 10000) > 0 && !quit)
    {
        if (fds[0].revents & POLLIN) // 检查实际发生了的事件
        {
            struct v4l2_buffer v4l2_buf;

            /* 输出相机数据 */
            memset(&v4l2_buf, 0, sizeof(v4l2_buf));
            v4l2_buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE; // 单平面模式
            if (ctx->capture_dmabuf)
                v4l2_buf.memory = V4L2_MEMORY_DMABUF; // 使用应用程序传入的缓冲区
            else
                v4l2_buf.memory = V4L2_MEMORY_MMAP; // 设备分配缓存

#if 1
            /*
             * 写入文件
             */
            save_to_jpg(ctx, &v4l2_buf);
#else
            /* 数据放入缓存队列，读取帧 */
            if (ioctl(ctx->cam_fd, VIDIOC_DQBUF, &v4l2_buf) < 0)
                ERROR_RETURN("无法从相机读取数据: %s (%d)", strerror(errno), errno);
#endif
            ctx->frame++;

            /* 将第n帧保存到文件 */
            if (ctx->frame == ctx->save_n_frame)
            {
                save_frame_to_file(ctx, &v4l2_buf);
            }

            if (ctx->cam_pixfmt == V4L2_PIX_FMT_MJPEG)
            {
                int fd = 0;
                uint32_t width, height, pixfmt;
                unsigned int i = 0;
                unsigned int eos_search_size = MJPEG_EOS_SEARCH_SIZE;//
                unsigned int bytesused = v4l2_buf.bytesused;// 读取帧的实际大小
                uint8_t *p;

                /* v4l2_buf.bytesused可能具有填充字节以进行对齐搜索EOF以获取确切大小 */
                if (eos_search_size > bytesused)
                    eos_search_size = bytesused;
                /* 查找图像尾部的 0xff 0xd9 */
                for (i = 0; i < eos_search_size; i++) {
                    p =(uint8_t *)(ctx->g_buff[v4l2_buf.index].start + bytesused);
                    if ((*(p-2) == 0xff) && (*(p-1) == 0xd9)) {
                        break;
                    }
                    bytesused--;
                }

                /* 解码MJPEG框架 */
                if (ctx->jpegdec->decodeToFd(fd, ctx->g_buff[v4l2_buf.index].start,
                    bytesused, pixfmt, width, height) < 0) // 将JPEG图像解码到硬件缓冲存储器
                    ERROR_RETURN("不能解码MJPEG");

                /* 将解码的缓冲区转换为YUV420P */
                if (-1 == NvBufferTransform(fd, ctx->render_dmabuf_fd, &transParams)) // 将一个DMA缓冲区转换为另一个DMA缓冲区
                    ERROR_RETURN("转换缓冲区失败");
            }
            else
            {
                if (ctx->capture_dmabuf)
                {
                    /* VIC操作的缓存同步，因为数据来自CPU */
                    NvBufferMemSyncForDevice(ctx->g_buff[v4l2_buf.index].dmabuff_fd, 0,
                            (void**)&ctx->g_buff[v4l2_buf.index].start);// 同步设备的硬件内存缓存
                }
                else
                {
                    /* 将原始缓冲区平面内容复制到NvBuffer平面 */
                    Raw2NvBuffer(ctx->g_buff[v4l2_buf.index].start, 0, ctx->cam_w, ctx->cam_h,
                                ctx->g_buff[v4l2_buf.index].dmabuff_fd);// 将原始缓冲区平面的内容复制到NvBuffer平面
                }

                /*  将相机缓冲区从YUV422转换为YUV420P */
                if (-1 == NvBufferTransform(ctx->g_buff[v4l2_buf.index].dmabuff_fd, ctx->render_dmabuf_fd,
                            &transParams)) // 将一个DMA缓冲区转换为另一个DMA缓冲区
                    ERROR_RETURN("转换缓冲区失败");

                if (ctx->cam_pixfmt == V4L2_PIX_FMT_GREY) {
                    if(!nvbuff_do_clearchroma(ctx->render_dmabuf_fd))
                        ERROR_RETURN("无法清除色度");
                }
            }
            cuda_postprocess(ctx, ctx->render_dmabuf_fd);

            /* 渲染缓冲区 */
            ctx->renderer->render(ctx->render_dmabuf_fd);

            /* 将相机缓冲区放回驱动程序 */
            if (ioctl(ctx->cam_fd, VIDIOC_QBUF, &v4l2_buf))
            {
                ERROR_RETURN("无法将相机缓冲区排队: %s (%d)", strerror(errno), errno);
            }
        }
    }

    /* 流停止时打印配置文件信息 */
    ctx->renderer->printProfilingStats();

    if (ctx->cam_pixfmt == V4L2_PIX_FMT_MJPEG)
        delete ctx->jpegdec;

    return true;
}

static bool
stop_stream(context_t * ctx)
{
    enum v4l2_buf_type type;

    /* Stop v4l2 streaming */
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(ctx->cam_fd, VIDIOC_STREAMOFF, &type))
        ERROR_RETURN("Failed to stop streaming: %s (%d)",
                strerror(errno), errno);

    INFO("Camera video streaming off ...");
    return true;
}

int
main(int argc, char *argv[])
{
    context_t ctx;
    int error = 0;

    /* 设置摄像头上下文默认参数 */
    set_defaults(&ctx);

    /* 解析参数及打印信息 */
    CHECK_ERROR(parse_cmdline(&ctx, argc, argv), cleanup, "指定的无效选项");

    /* 初始化相机和EGL显示,EGL显示将被用来映射缓冲区到CUDA缓冲for CUDA处理 */
    CHECK_ERROR(init_components(&ctx), cleanup, "未能初始化v4l2组件");

    if (ctx.cam_pixfmt == V4L2_PIX_FMT_MJPEG)
    {
        CHECK_ERROR(prepare_buffers_mjpeg(&ctx), cleanup, "没有准备v4l2 buffs");
    }
    else
    {
        CHECK_ERROR(prepare_buffers(&ctx), cleanup, "没有准备v4l2 buffs");
    }

    CHECK_ERROR(start_stream(&ctx), cleanup, "未能开始流");

    CHECK_ERROR(start_capture(&ctx), cleanup, "没有开始捕捉")

    CHECK_ERROR(stop_stream(&ctx), cleanup, "停止流");

cleanup:
    if (ctx.cam_fd > 0)
        close(ctx.cam_fd);

    if (ctx.renderer != NULL)
        delete ctx.renderer;

    if (ctx.egl_display && !eglTerminate(ctx.egl_display))
        printf("Failed to terminate EGL display connection\n");

    if (ctx.g_buff != NULL)
    {
        for (unsigned i = 0; i < V4L2_BUFFERS_NUM; i++) {
            if (ctx.g_buff[i].dmabuff_fd)
                NvBufferDestroy(ctx.g_buff[i].dmabuff_fd);
            if (ctx.cam_pixfmt == V4L2_PIX_FMT_MJPEG)
                munmap(ctx.g_buff[i].start, ctx.g_buff[i].size);
        }
        free(ctx.g_buff);
    }

    NvBufferDestroy(ctx.render_dmabuf_fd);

    if (error)
        printf("App run failed\n");
    else
        printf("App run was successful\n");

    return -error;
}
